from typing import Iterable, Optional, Protocol, TypeAlias

from syntactes import Grammar, Token
from syntactes._action import Action
from syntactes._state import LR0State, LR1State, State
from syntactes.parsing_table import Conflict, Entry

Row: TypeAlias = dict[Token, list[Action]]


class ParsingTable(Protocol):
    rows: dict[State, Row]
    initial_state: State

    @staticmethod
    def from_entries(entries: Iterable[Entry], grammar: Grammar) -> "ParsingTable": ...

    def get(self, state: State) -> Optional[Row]: ...

    def get_actions(self, state: State, token: Token) -> Optional[list[Action]]: ...

class LR0ParsingTable:
    """
    Table that contains all the transitions from state to state with a symbol.
    """

    def __init__(self, grammar: Grammar) -> None:
        self.rows: dict[LR0State, Row] = dict()
        self._grammar = grammar
        self._initial_state = None

    @staticmethod
    def from_entries(entries: Iterable[Entry], grammar: Grammar) -> "LR0ParsingTable":
        """
        Create a parsing table from the given entries.
        """
        table = LR0ParsingTable(grammar)
        {table.add_entry(entry) for entry in entries}
        return table

    @property
    def initial_state(self) -> LR0State:
        return self._initial_state

    def get_actions(self, state: LR0State, token: Token) -> Optional[list[Action]]:
        """
        Get the actions from state with given number with `token`.
        If there are no actions, returns None.
        """
        return self.rows.get(state, {}).get(token, None)

    def get(self, state: LR0State) -> Optional[Row]:
        """
        Get the mapping of tokens to actions for the given state number.
        Returns None if the state is not found.
        """
        return self.rows.get(state, None)

    def add_entry(self, entry: Entry) -> None:
        """
        Adds an entry to the parsing table.
        """
        if entry.from_state.number == 1:
            self._initial_state = entry.from_state

        row = self.rows.setdefault(entry.from_state, {})
        actions = row.setdefault(entry.token, list())
        actions.append(entry.action)

    def pretty_str(self) -> str:
        """
        Returns a pretty-formatted string representation of the table.
        """
        return self._rules_pretty_str() + "\n\n" + self._table_pretty_str()

    def conflicts(self) -> list[Conflict]:
        """
        Retuns a list with all the conflicts in the parsing table.
        """
        conflicts = []
        for state, row in self.rows.items():
            for token, actions in row.items():
                if len(actions) > 1:
                    conflicts.append(Conflict(state, token, actions))

        return conflicts

    def _rules_pretty_str(self) -> str:
        rules = [str(i) + ". " + str(r) for i, r in enumerate(self._grammar.rules)]
        rules_str = "\n".join(rules)
        rules_str = "GRAMMAR RULES\n" + "-" * max(map(len, rules)) + "\n" + rules_str
        rules_str += "\n" + "-" * max(map(len, rules))

        return rules_str

    def _table_pretty_str(self) -> str:
        rows = []
        tokens = sorted(self._grammar.tokens)
        for number, row in sorted(
            map(lambda tpl: (tpl[0].number, tpl[1]), self.rows.items())
        ):
            r = [str(number)]
            for token in sorted(tokens):
                actions = row.get(token, [])
                if len(actions) >= 1:
                    actions_str = ",".join(map(str, actions))
                else:
                    actions_str = "--"

                r.append(actions_str)

            rows.append(r)

        table = "|     |  "
        table += "   |  ".join(str(token) for token in tokens) + "  |\n"

        header = self._header_str() + "\n" + "-" * len(table) + "\n"
        table += "-" * len(table) + "\n"

        for row in rows:
            new_row = "|  " + "  |  ".join(row) + " |" + "\n"
            table += new_row
            table += "-" * len(new_row) + "\n"

        return header + table

    def _header_str(self) -> str:
        return "LR0 PARSING TABLE"


class SLRParsingTable(LR0ParsingTable):
    @staticmethod
    def from_entries(
        entries: Iterable[Entry], tokens: Iterable[Token]
    ) -> "SLRParsingTable":
        """
        Create a parsing table from the given entries.
        """
        table = SLRParsingTable(tokens)
        {table.add_entry(entry) for entry in entries}
        return table

    def _header_str(self) -> str:
        return "SLR PARSING TABLE"


class LR1ParsingTable(LR0ParsingTable):
    """
    Table that contains all the transitions from state to state with a symbol.
    """

    def __init__(self, grammar: Grammar) -> None:
        self.rows: dict[LR1State, Row] = dict()
        self._grammar = grammar
        self._initial_state = None

    @staticmethod
    def from_entries(
        entries: Iterable[Entry], tokens: Iterable[Token]
    ) -> "LR1ParsingTable":
        """
        Create a parsing table from the given entries.
        """
        table = LR1ParsingTable(tokens)
        {table.add_entry(entry) for entry in entries}
        return table

    @property
    def initial_state(self) -> LR1State:
        return self._initial_state

    def get_actions(self, state: LR1State, token: Token) -> Optional[list[Action]]:
        """
        Get the actions from state with given number with `token`.
        If there are no actions, returns None.
        """
        return self.rows.get(state, {}).get(token, None)

    def get(self, state: LR1State) -> Optional[Row]:
        """
        Get the mapping of tokens to actions for the given state number.
        Returns None if the state is not found.
        """
        return self.rows.get(state, None)

    def _header_str(self) -> str:
        return "LR1 PARSING TABLE"



from typing import TypeAlias

from syntactes import Token
from syntactes._action import Action
from syntactes._state import LR0State

Row: TypeAlias = dict[Token, list[Action]]


class Entry:
    """
    An entry of the parsing table. Holds the information of a transition from
    a state to another state via a symbol.
    """

    def __init__(self, from_state: LR0State, token: Token, action: Action) -> None:
        self.from_state = from_state
        self.token = token
        self.action = action

    def __repr__(self) -> str:
        return f"<Entry: {str(self)}>"

    def __str__(self) -> str:
        return f"{self.from_state.number}, {self.action}, {self.token}"

    def __hash__(self) -> int:
        return hash((self.from_state, self.token, self.action))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Entry):
            return False

        return (
            self.from_state == other.from_state
            and self.token == other.token
            and self.action == other.action
        )



from enum import Enum

from syntactes import Token
from syntactes._action import Action, ActionType
from syntactes._state import LR0State


class ConflictType(Enum):
    SHIFT_SHIFT = "shift/shift"
    SHIFT_REDUCE = "shift/reduce"
    REDUCE_REDUCE = "reduce/reduce"
    NO_CONFLICT = "NO_CONFLICT"


class Conflict:
    def __init__(self, state: LR0State, token: Token, actions: list[Action]) -> None:
        self.state = state
        self.token = token
        self.actions = actions
        self._conflict_type = None

    def pretty_str(self) -> str:
        """
        Returns a pretty-formatted string with conflict details.
        """
        string = f"{self.conflict_type.value} conflict in state {self.state}\n"
        string += f"Available actions on input token '{self.token}':\n"
        for action in self.actions:
            if action.action_type == ActionType.SHIFT:
                string += f"shift to {action.actionable}\n"
            elif action.action_type == ActionType.REDUCE:
                string += f"reduce by {action.actionable}\n"
            else:
                string += f"{action.action_type}, {action.actionable}\n"

        return string

    @property
    def conflict_type(self) -> ConflictType:
        if self._conflict_type is None:
            self._conflict_type = self._create_type()

        return self._conflict_type

    def _create_type(self) -> ConflictType:
        if not len(self.actions) > 1:
            return ConflictType.NO_CONFLICT

        action_types = set(map(lambda a: a.action_type, self.actions))

        if ActionType.SHIFT not in action_types:
            if ActionType.REDUCE not in action_types:
                return ConflictType.NO_CONFLICT

            return ConflictType.REDUCE_REDUCE

        if ActionType.REDUCE not in action_types:
            return ConflictType.SHIFT_SHIFT

        return ConflictType.SHIFT_REDUCE

    def __repr__(self) -> str:
        return f"<Conflict: {str(self)}>"

    def __str__(self) -> str:
        return f"{self.state}, {self.token}, {self.actions}"



from abc import ABC, abstractmethod
from typing import Optional, Type

from syntactes import Grammar, Token
from syntactes._action import Action
from syntactes._item import Item, LR0Item, LR1Item
from syntactes._state import LR0State, LR1State, State
from syntactes.parsing_table import (
    Entry,
    LR0ParsingTable,
    LR1ParsingTable,
    ParsingTable,
    SLRParsingTable,
)


class Generator(ABC):
    table_cls: Type[ParsingTable]
    state_cls: Type[State]
    item_cls: Type[Item]

    def __init__(self, grammar: Grammar) -> None:
        self.grammar = grammar

    @abstractmethod
    def closure(self, items: set[Item]) -> set[Item]:
        raise NotImplementedError()

    @abstractmethod
    def goto(self, items: set[Item], token: Token) -> set[Item]:
        raise NotImplementedError()

    def generate(self) -> ParsingTable:
        """
        Generates an parsing table for the configured grammar.
        """
        states, shift_entries = self._create_states_and_shift_entries()
        reduce_entries = self._create_reduce_entries(states)

        entries = shift_entries | reduce_entries

        table = self.table_cls.from_entries(entries, self.grammar)

        return table

    def get_states(self) -> set[State]:
        """
        Returns the set of automaton states for the configured grammar.
        """
        states, _ = self._create_states_and_shift_entries()
        return states

    def _first(self, *symbols) -> set[Token]:
        """
        Computes and returns the FIRST set for the given symbols.

        The FIRST set of a symbol 'G' is the set of terminal symbols that are
        first in the right-hand side of a rule where 'G' is the left-hand side.

        e.g. 't', 'k' and 'a' would be the FIRST set of G for the below rules:
        1. G -> t
        2. G -> kM
        3. G -> T
        4. T -> a
        where M is either terminal or non-terminal and T is non-terminal.
        'a' would be included in the FIRST set because if rule 4 is substituted in
        rule 3, 'a' (which is a terminal) could be derived from 'G'.

        The computation of the FIRST set looks very simple if symbols = X Y Z, it seems
        as if Y and Z can be ignored and FIRST(X) is the only thing that matters.
        But consider a grammar where X -> Y and Y -> ε. Because Y can produce the empty
        string - and therefore X can produce the empty string - we find that FIRST(XYZ)
        must include FIRST(Z). Therefore, in computing FIRST sets we must keep track of
        which symbols can produce the empty string.
        """
        if len(symbols) == 0:
            return set()

        symbol = symbols[0]

        if symbol.is_terminal:
            return {symbol}

        _set: set[Token] = set()

        for rule in self.grammar.rules:
            if rule.lhs != symbol:
                continue

            if rule.has_null_rhs() and len(symbols) > 1:
                _set |= self._first(symbols[1:])
                continue

            if rule.rhs[0].is_terminal:
                _set.add(rule.rhs[0])
            elif rule.rhs_len == 1:
                _set |= self._first(*rule.rhs)

        return _set

    def _follow(self, symbol: Token) -> set[Token]:
        """
        Computes and returns the FOLLOW set for the given symbol.

        The FOLLOW set of a symbol 'G' is the set of terminals that can immediately
        follow 'G' in a rule.
        """
        if symbol.is_terminal:
            return set()

        _set: set[Token] = set()

        for rule in self.grammar.rules:
            for i, s in enumerate(rule.rhs):
                if s != symbol:
                    continue

                if i == rule.rhs_len - 1:
                    if rule.lhs == symbol:
                        continue

                    _set |= self._follow(rule.lhs)
                else:
                    _set |= self._first(rule.rhs[i + 1])

        return _set

    def _create_states_and_shift_entries(self) -> tuple[set[State], set[Entry]]:
        """
        Computes and returns the states and entries for shift actions.
        """
        states, entries = dict(), set()

        initial_items = self._create_initial_items()
        initial_state = self.state_cls.from_items(initial_items)
        initial_state.set_number(1)
        states[initial_state] = 1

        _states, _entries = dict(), set()
        while (_states, _entries) != (states, entries):
            _states = {s: n for s, n in states.items()}
            _entries = {e for e in entries}
            states, entries = self._extend_states_and_shift_entries(_states, _entries)

        return set(states.keys()), entries

    def _extend_states_and_shift_entries(
        self, states: dict[State, int], entries: set[Entry]
    ) -> tuple[dict[State, int], set[Entry]]:
        """
        Extends states and entries following the below algorithm:

        ```
        for each state S in states
            for each item A -> a.Xb in S
                J = goto(S, X)
                states.add(J)
                entries.add((S->J, X))
        ```
        """
        _states = {s: n for s, n in states.items()}
        _entries = {e for e in entries}

        EOF = Token.eof()
        for state in states:
            for item in state.items:
                if item.dot_is_last():
                    continue

                if item.after_dot == EOF:
                    state.set_final()
                    continue

                new_items = self.goto(state.items, item.after_dot)

                if len(new_items) == 0:
                    continue

                new = self.state_cls.from_items(new_items)

                number = _states.setdefault(new, len(_states) + 1)
                new.set_number(number)

                _entries.add(Entry(state, item.after_dot, Action.shift(new)))

        return _states, _entries

    @abstractmethod
    def _create_initial_items(self) -> set[Item]:
        raise NotImplementedError()

    @abstractmethod
    def _create_reduce_entries(self, states: set[State]) -> set[Entry]:
        raise NotImplementedError()


class LR0Generator(Generator):
    """
    Generator of LR0 parsing tables.
    """

    table_cls = LR0ParsingTable
    state_cls = LR0State
    item_cls = LR0Item

    def closure(self, items: set[LR0Item]) -> set[LR0Item]:
        """
        Computes and returns the closure for the given set of items.

        The closure operation adds more items to a set of items when there
        is a dot to the left of a non-terminal symbol.

        e.g.
        for any item S -> . E in the given items, closure adds E -> . T
        and T -> . x, where E -> T and T -> x are production rules.
        """
        _set = {item for item in items}
        __set = set()

        while __set != _set:
            __set = {i for i in _set}

            for item in items:
                if item.dot_is_last():
                    continue

                new_items = self._get_related_items(item.after_dot)
                _set |= new_items

        return _set

    def goto(self, items: set[LR0Item], token: Token) -> set[LR0Item]:
        """
        Computes and returns the GOTO set for the given set of items.

        The goto operation creates a set where all items have the dot past the
        given symbol.
        """
        _set: set[LR0Item] = set()

        for item in items:
            if item.dot_is_last() or item.after_dot != token:
                continue

            next_item = LR0Item(item.rule, item.position + 1)
            _set.add(next_item)

        return self.closure(_set)

    def _get_related_items(self, symbol: Token) -> set[LR0Item]:
        """
        e.g. the items X -> .g, Y -> .p would be returned for the below grammar rules:
        1. X -> g
        2. X -> Y
        3. Y -> p
        where 'g' and 'p' are terminals.
        """
        _set: set[LR0Item] = set()

        for rule in self.grammar.rules:
            if rule.lhs == symbol:
                _set.add(LR0Item(rule, 0))

                if rule.rhs_len == 1 and not rule.rhs[0].is_terminal:
                    _set |= self._get_related_items(rule.rhs[0])

        return _set

    def _create_initial_items(self) -> set[LR0Item]:
        return self.closure({LR0Item(self.grammar.starting_rule, 0)})

    def _create_reduce_entries(self, states: set[LR0State]) -> set[Entry]:
        """
        Computes and returns the entries for reduce actions and the accept action.
        """
        entries: set[Entry] = set()

        for state in states:
            for item in state.items:
                if item.after_dot == Token.eof():
                    entries.add(Entry(state, Token.eof(), Action.accept()))

                if not item.dot_is_last():
                    continue

                for token in self.grammar.tokens:
                    if token.is_terminal:
                        entries.add(Entry(state, token, Action.reduce(item.rule)))

        return entries


class SLRGenerator(LR0Generator):
    table_cls = SLRParsingTable

    def _create_reduce_entries(self, states: set[LR0State]) -> set[Entry]:
        """
        Computes and returns the entries for reduce actions and the accept action.
        """
        entries: set[Entry] = set()

        for state in states:
            for item in state.items:
                if item.after_dot == Token.eof():
                    entries.add(Entry(state, Token.eof(), Action.accept()))

                if not item.dot_is_last():
                    continue

                for token in self._follow(item.rule.lhs):
                    entries.add(Entry(state, token, Action.reduce(item.rule)))

        return entries


class LR1Generator(Generator):
    table_cls = LR1ParsingTable
    state_cls = LR1State
    item_cls = LR1Item

    def closure(self, items: set[LR1Item]) -> set[LR1Item]:
        """
        Computes and returns the closure for the given set of items.

        The closure operation adds more items to a set of items when there
        is a dot to the left of a non-terminal symbol.
        """
        _set = {item for item in items}
        __set = set()

        while __set != _set:
            __set = {i for i in _set}

            for item in __set:
                if item.dot_is_last():
                    continue

                if item.position + 1 < item.rule.rhs_len:
                    next_symbol = item.rule.rhs[item.position + 1]
                else:
                    next_symbol = None

                new_items = self._get_related_items(
                    item.after_dot, next_symbol, item.lookahead_token
                )
                _set |= new_items

        return _set

    def goto(self, items: set[LR1Item], token: Token) -> set[LR1Item]:
        """
        Computes and returns the GOTO set for the given set of items.

        The goto operation creates a set where all items have the dot past the
        given symbol.
        """
        _set: set[LR1Item] = set()

        for item in items:
            if item.dot_is_last() or item.after_dot != token:
                continue

            next_item = LR1Item(item.rule, item.position + 1, item.lookahead_token)
            _set.add(next_item)

        return self.closure(_set)

    def _get_related_items(
        self, symbol: Token, next_symbol: Optional[Token], lookahead_token: Token
    ) -> set[LR1Item]:
        _set: set[LR1Item] = set()

        if next_symbol is None:
            lookaheads = (lookahead_token,)
        else:
            lookaheads = (next_symbol, lookahead_token)

        for rule in self.grammar.rules:
            if rule.lhs != symbol:
                continue

            for s in self._first(*lookaheads):
                _set.add(LR1Item(rule, 0, s))

                if rule.rhs_len == 1 and not rule.rhs[0].is_terminal:
                    _set |= self._get_related_items(rule.rhs[0], None, s)

        return _set

    def _create_reduce_entries(self, states: set[LR1State]) -> set[Entry]:
        """
        Computes and returns the entries for reduce actions and the accept action.
        """
        entries: set[Entry] = set()

        for state in states:
            for item in state.items:
                if item.after_dot == Token.eof():
                    entries.add(Entry(state, Token.eof(), Action.accept()))

                if not item.dot_is_last():
                    continue

                entries.add(
                    Entry(state, item.lookahead_token, Action.reduce(item.rule))
                )

        return entries

    def _create_initial_items(self) -> set[LR1Item]:
        return self.closure({LR1Item(self.grammar.starting_rule, 0, Token.eof())})



class Token:
    """
    A token of the grammar. Can be a terminal or non-terminal symbol.
    """

    def __init__(self, symbol: str, is_terminal: bool, value=None) -> None:
        self.symbol = symbol
        self.is_terminal = is_terminal
        self.value = value

    @staticmethod
    def null() -> "Token":
        """
        Returns the NULL token.
        """
        return Token("ε", True)

    @staticmethod
    def eof() -> "Token":
        """
        Returns the EOF token.
        """
        return Token("$", True)

    def __repr__(self) -> str:
        return f"<Token: {self}>"

    def __str__(self) -> str:
        return self.symbol

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Token):
            return False

        return self.symbol == other.symbol and self.is_terminal is other.is_terminal

    def __lt__(self, other) -> bool:
        if not isinstance(other, Token):
            raise ValueError(
                f"'<' not supported between instances of 'Token' and {type(other).__name__}"
            )

        return self.symbol < other.symbol



class ParserError(Exception): ...


class UnexpectedTokenError(ParserError):
    """
    A token was received that does not map to an action. The stream of tokens
    is syntactically invalid.
    """

    def __init__(self, received_token, expected_tokens):
        self.received_token = received_token
        self.expected_tokens = expected_tokens
        msg = f"Received token: {received_token}; expected one of: {[str(e) for e in expected_tokens]}"
        super().__init__(msg)


class NotAcceptedError(ParserError):
    """
    The parser did not receive an accept action. The stream of tokens is
    syntactically invalid.
    """



from abc import ABC
from collections import deque
from typing import Iterable, Type

from syntactes import Grammar, LR0Generator, LR1Generator, SLRGenerator, Token
from syntactes._action import Action, ActionType
from syntactes._state import LR0State
from syntactes.parser import (
    ExecutablesRegistry,
    NotAcceptedError,
    ParserError,
    UnexpectedTokenError,
)
from syntactes.parsing_table import ParsingTable


class Parser(ABC):
    generator_cls: Type

    def __init__(self, table: ParsingTable) -> None:
        self._table = table
        self._token_stack: deque[Token] = deque()
        self._state_stack: deque[LR0State] = deque()
        self._token_stream: deque[Token] = deque()

    @classmethod
    def from_grammar(cls, grammar: Grammar) -> "Parser":
        """
        Create a parser for the given grammar.
        """
        generator = cls.generator_cls(grammar)
        parsing_table = generator.generate()
        parser = cls(parsing_table)
        return parser

    def parse(self, stream: Iterable[Token]) -> None:
        """
        Parses the given stream of tokens. Expects the EOF token as the last one.

        Raises `syntactes.parser.UnexpectedTokenError` if an unexpected token is
        received.

        Raises `syntactes.parser.NotAcceptedError` if the stream of token has been
        parsed and the parser did not receive an accept action.
        """
        self._set_state(self._table.initial_state)
        self._token_stream.extend(stream)

        while len(self._token_stream) > 0:
            token = self._token_stream.popleft()
            self._apply_action(token, self._get_action(token))

        if token != Token.eof():
            self._raise(NotAcceptedError("Expected EOF token. "))

        if not self._get_state().is_final:
            actions = self._table.get(self._get_state())
            expected_tokens = [] if actions is None else list(actions.keys())
            self._raise(UnexpectedTokenError(Token.eof(), expected_tokens))

    def _apply_action(self, token: Token, action: Action) -> None:
        if action.action_type == ActionType.SHIFT:
            self._token_stack.append(token)
            self._set_state(action.actionable)
        elif action.action_type == ActionType.REDUCE:
            rule = action.actionable
            args = [self._token_stack.pop() for _ in reversed(rule.rhs)]
            self._token_stack.append(rule.lhs)

            {self._state_stack.pop() for _ in rule.rhs}

            executable = ExecutablesRegistry.get(rule)
            executable(*args)

            self._token_stream.appendleft(token)  # reduce actions do not consume tokenA

            shift = self._get_action(rule.lhs)
            self._set_state(shift.actionable)

    def _get_action(self, token: Token) -> Action:
        actions = self._table.get_actions(self._get_state(), token)
        if actions is None:
            actions = self._table.get(self._get_state())
            expected_tokens = [] if actions is None else list(actions.keys())
            self._raise(UnexpectedTokenError(token, expected_tokens))

        action = self._resolve_conflict(actions)
        return action

    def _resolve_conflict(self, actions: list[Action]) -> Action:
        return actions[0]

    def _set_state(self, state: LR0State) -> None:
        self._state_stack.append(state)

    def _get_state(self) -> LR0State:
        return self._state_stack[-1]

    def _cleanup(self) -> None:
        self._token_stack.clear()
        self._state_stack.clear()
        self._token_stream.clear()

    def _raise(self, error: ParserError) -> None:
        self._cleanup()
        raise error from None


class LR0Parser(Parser):
    generator_cls = LR0Generator


class SLRParser(Parser):
    generator_cls = SLRGenerator


class LR1Parser(Parser):
    generator_cls = LR1Generator



import functools
from collections.abc import Callable
from typing import TypeAlias

from syntactes import Rule

Executable: TypeAlias = Callable[[...], None]


def execute_on(rule: Rule):
    """
    Decorate a function to be executed upon recognition of `rule` by the parser.
    """

    def executable_decorator(executable_fn: Executable) -> Executable:
        ExecutablesRegistry.register(rule, executable_fn)

        @functools.wraps(executable_fn)
        def wrapped_executable_fn(*args, **kwargs) -> None:
            return executable_fn(*args, **kwargs)

        return wrapped_executable_fn

    return executable_decorator


class ExecutablesRegistry:
    """
    Registry of executable functions, i.e. functions that get called when a grammar
    rule is recognized by the parser.
    """

    _registry: dict[Rule, Executable] = {}

    @classmethod
    def register(cls, rule: Rule, executable_fn: Executable) -> None:
        """
        Register a function to be executed upon recognition of the given rule.
        """
        cls._registry[rule] = executable_fn

    @classmethod
    def get(cls, rule: Rule) -> Executable:
        """
        Get the executable registered for the given rule.
        If no executable is registered returns a function that does nothing.
        """
        return cls._registry.get(rule, lambda *_, **__: None)

    @classmethod
    def clear(cls) -> None:
        """
        Clear all registered rules.
        """
        cls._registry.clear()
